from flask import url_for


def mapper_index():
    return "Mapper Index Page"


def overridden_endpoint():
    return "The url for url_for('endpoint_name') is " + url_for("endpoint_name")


def post_only():
    return "Only accessible via POST"



from flask import Blueprint, Markup, request, render_template
from os import path
from pkgutil import walk_packages
from werkzeug.utils import import_string
from werkzeug.exceptions import default_exceptions, HTTPException


def autoload(app, apps_package="apps", module_name="routes", blueprint_name="routes", on_error=None):
    """Automatically load Blueprints from the specified package and registers them with Flask."""

    if not apps_package:
        raise ValueError("No apps package provided - unable to begin autoload")

    if isinstance(apps_package, basestring):
        package_code = import_string(apps_package)
    else:
        #: `apps_package` can be the already imported parent package
        #: (i.e. the following is a licit pattern)::
        #:
        #:      import app_package
        #:      # do something else with app_package
        #:      autoload(app, app_package)
        package_code = apps_package
        apps_package = apps_package.__name__

    package_paths = package_code.__path__

    package_paths = [path.join(app.config["ROOT_PATH"], p) for p in package_paths]
    apps_package = apps_package + u"." if not apps_package.endswith(".") else apps_package

    if on_error is None:
        on_error = lambda name: app.logger.warn("Unable to import {name}.".format(name=name))

    import_template = "{base}.{module}.{symbol}"

    #: Autoloaded apps must be Python packages
    for sub_app_name in walk_packages(path=package_paths, prefix=apps_package, onerror=on_error):

        _to_import = import_template.format(base=sub_app_name, \
                                                module=module_name, \
                                                symbol=blueprint_name)
        sub_app = import_string(_to_import)

        if isinstance(sub_app, Blueprint):
            app.register_blueprint(sub_app)
        else:
            app.logger.warn(("Failed to register {name} - "
                "it does not match the registration pattern.").format(name=sub_app_name))


def setup_errors(app, error_template="errors.html"):
    def error_handler(error):
        if isinstance(error, HTTPException):
            description = error.get_description(request.environ)
            code = error.code
            name = error.name
        else:
            description = error
            code = 500
            name = "Internal Server Error"
        return render_template(error_template,
                                code=code,
                                name=Markup(name),
                                description=Markup(description))

    for exception in default_exceptions:
        app.register_error_handler(exception, error_handler)



import unittest

from flask import Flask
from hip_pocket import Mapper


class MapperAppTests(unittest.TestCase):
    def setUp(self):
        self.app = Flask("mapper")
        self.app.debug = True
        self.app.testing = True
        self.mapper = Mapper(self.app)

    def test_endpoint_setup(self):
        self.assertEqual(self.mapper.endpoint_base_name, "mapper")

    def test_mapping_creates_route(self):
        self.mapper.add_url_rule("/", "mapper_index")
        routes = list(self.app.url_map.iter_rules())
        # Flask adds the static route for us automatically
        self.assertEqual(len(routes), 2)

    def test_import_prefixed(self):
        self.mapper.add_url_rule("/", "mapper_index")
        routes = list(self.app.url_map.iter_rules())
        our_route = [route for route in routes if route.rule == "/"][0]
        self.assertEqual(our_route.endpoint, "mapper.mapper_index")

    def test_import_resolves(self):
        self.mapper.add_url_rule("/", "mapper_index")
        with self.app.test_client() as tc:
            rv = tc.get("/")
            self.assertTrue("Mapper Index Page" in rv.data)

    def test_overriding_endpoint(self):
        self.mapper.add_url_rule("/", "overridden_endpoint", endpoint="endpoint_name")
        with self.app.test_client() as tc:
            rv = tc.get("/")
            self.assertTrue("The url for url_for('endpoint_name') is /" in rv.data)

    def test_url_kwarg_forwarding(self):
        self.mapper.add_url_rule("/", "post_only", methods=["POST"])
        with self.app.test_client() as tc:
            self.assertTrue("405" in tc.get("/").status)
            self.assertTrue("Only accessible via POST" in tc.post("/").data)


def suite():
    suite = unittest.TestLoader().loadTestsFromTestCase(MapperAppTests)
    return suite



from functools import wraps
from werkzeug import import_string, cached_property

from .tasks import autoload, setup_errors

__all__ = ["HipPocket", "LateLoader", "Mapper"]


class HipPocket(object):
    """A wrapper around Flask to help make managing complex applications easier.

    Why not use Flask's `flask.Flask.before_first_request` decorator?  Because sometimes
    you need a reference to the current application to do your setup - and it's easier to
    do it with a wrapper.

    Alternately, because a flask is always better when you have something to carry it in."""

    def __init__(self, app=None, tasks=None):
        """
        :param app: A `flask.Flask` instance.  If none is provided the :meth: init_app  method may be called later
        to initialize the `HipPocket` instance.
        :param tasks: An iterable of callable objects. Each callable will be passed the app in turn.
        """

        self.tasks = tasks if tasks is not None else [autoload, setup_errors]
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        """Runs each of the tasks added via the `task` decorator."""

        for task in self.tasks:
            task(app)

    def task(self, func, *args, **kwargs):
        """A task is a unit of work to be performed by HipPocket on an initialized Flask app.

        All tasks must take the initialized app as their first argument.
        They can take any other arguments at all - these arguments must be provided to the
        `@task` decorator during decoration.

        Examples::

            hip_pocket = HipPocket()

            @hip_pocket.task
            def do_something(app):
                # Do something with app

            @hip_pocket.task(an_arg, another_arg)
            def do_something_else(app, an_arg, another_arg):
                # Do something with app, an_arg and another_arg

            app = Flask(__name__)

            hip_pocket.init_app(app)

        See `tasks.py` for some examples tasks.

        NOTE: The task decorator does not handle single callable arguments. If you need
        to pass single callables include a nonsense keyword argument.

            # This will fail
            @hip_pocket.task(lambda x: x*2)
            def schedule_doubles(app, manager):
                # Have the manager schedule something.

            # This will work as anticipated
            @hip_pocket.task(lambda x: x**2, doubles=2):
            def add_pocket_squares(app, butler, **kwargs):
                # Add pocket squares to the appropriate garments.
        """
        def decorator(f):
            @wraps(f)
            def inner(app):
                return f(app, *args, **kwargs)
            return inner

        if not callable(func) or len(args) > 0 or len(kwargs) > 0:
            def wrapper(func):
                task = decorator(func)
                self.tasks.append(task)
                return task
            return wrapper
        else:
            task = decorator(func)
            self.tasks.append(task)
            return task


class LateLoader(object):
    """Provides a way of loading views on the first request for the view
    rather than on application initialization.  This speeds up application load times
    and keeps resource requirements low (since you don't load code until you need it).

    Taken directly from the example in http://flask.pocoo.org/docs/patterns/lazyloading/"""
    def __init__(self, import_name):
        self.__module__, self.__name__ = import_name.rsplit(u".", 1)
        self.import_name = import_name

    @cached_property
    def view(self):
        return import_string(self.import_name)

    def __call__(self, *args, **kwargs):
        return self.view(*args, **kwargs)


class Mapper(object):
    """Provides a convenience wrapper around `LateLoader`.

    It is designed to be used as follows::

        from flask import Blueprint
        from hip_pocket import Mapper

        routes = Blueprint("my_blueprint", __name__, url_prefix="/my-blueprint")
        mapper = Mapper(routes, "apps_package.my_app")

        mapper.add_url_rule("/", "my_module.my_endpoint")

    `apps_package.my_app.my_module.my_endpoint` will be imported
    when the URL `/my-blueprint/` is hit for the first time.
    This import will be cached in order to ensure that
    subsequent requests to this url will **not** result in additional imports."""
    def __init__(self, blueprint, endpoint_base_name=None, **url_defaults):
        self.blueprint = blueprint
        self.url_defaults = url_defaults
        self.endpoint_base_name = endpoint_base_name \
                                    if endpoint_base_name is not None \
                                    else blueprint.import_name

    def add_url_rule(self, url, import_name, **url_kwargs):

        endpoint = u".".join([self.endpoint_base_name, import_name])
        view = LateLoader(endpoint)

        url_defaults = self.url_defaults.copy()
        # Allow overriding of endpoints for `url_for`.
        url_defaults["endpoint"] = endpoint
        url_defaults.update(**url_kwargs)
        url_kwargs = url_defaults

        self.blueprint.add_url_rule(url, view_func=view, **url_kwargs)



"""
**Unit** tests for Hip Pocket.
"""

import os
import sys
import unittest
from werkzeug.utils import import_string, find_modules

# The following are taken from Flask's unit test suite with much thanks to Armin
# https://github.com/mitsuhiko/flask/blob/master/flask/testsuite/__init__.py


def add_to_path(path):
    """Adds an entry to sys.path if it's not already there.  This does
    not append it but moves it to the front so that we can be sure it
    is loaded.
    """
    if not os.path.isdir(path):
        raise RuntimeError('Tried to add nonexisting path')

    def _samefile(x, y):
        try:
            return os.path.samefile(x, y)
        except (IOError, OSError):
            return False
    sys.path[:] = [x for x in sys.path if not _samefile(path, x)]
    sys.path.insert(0, path)


def iter_suites():
    """Yields all testsuites."""
    for module in find_modules(__name__):
        mod = import_string(module)
        if hasattr(mod, 'suite'):
            yield mod.suite()


def find_all_tests(suite):
    """Yields all the tests and their names from a given suite."""
    suites = [suite]
    while suites:
        s = suites.pop()
        try:
            suites.extend(s)
        except TypeError:
            yield s, '%s.%s.%s' % (
                s.__class__.__module__,
                s.__class__.__name__,
                s._testMethodName
            )


class BetterLoader(unittest.TestLoader):
    """A nicer loader that solves two problems.  First of all we are setting
    up tests from different sources and we're doing this programmatically
    which breaks the default loading logic so this is required anyways.
    Secondly this loader has a nicer interpolation for test names than the
    default one so you can just do ``run-tests.py ViewTestCase`` and it
    will work.
    """

    def getRootSuite(self):
        return suite()

    def loadTestsFromName(self, name, module=None):
        root = self.getRootSuite()
        if name == 'suite':
            return root

        all_tests = []
        for testcase, testname in find_all_tests(root):
            if testname == name or \
               testname.endswith('.' + name) or \
               ('.' + name + '.') in testname or \
               testname.startswith(name + '.'):
                all_tests.append(testcase)

        if not all_tests:
            raise LookupError('could not find test case for "%s"' % name)

        if len(all_tests) == 1:
            return all_tests[0]
        rv = unittest.TestSuite()
        for test in all_tests:
            rv.addTest(test)
        return rv


def setup_path():
    add_to_path(os.path.abspath(os.path.join(
        os.path.dirname(__file__), 'test_apps')))


def suite():
    """A testsuite that has all the Flask tests.  You can use this
    function to integrate the Flask tests into your own testsuite
    in case you want to test that monkeypatches to Flask do not
    break it.
    """
    setup_path()
    suite = unittest.TestSuite()
    for other_suite in iter_suites():
        suite.addTest(other_suite)
    return suite


def main():
    """Runs the testsuite as command line application."""
    try:
        unittest.main(testLoader=BetterLoader(), defaultTest='suite')
    except Exception, e:
        print 'Error: %s' % e


if __name__ == "__main__":
    unittest.main(testLoader=BetterLoader(), defaultTest='suite')



__all__ = ["__version__", "VERSION"]
__version__ = (0, 1, 1, "alpha")

maj, min, patch = [unicode(v) for v in __version__[:3]]
ver = u"{maj}.{min}.{patch}".format(maj=maj, min=min, patch=patch)
addenda = u".".join([unicode(v) for v in __version__[3:]])

#: HipPocket complies with http://semver.org/ v2.0.0-rc.1
VERSION = u"{ver}-{addenda}".format(ver=ver, addenda=addenda).strip(u"-")



from os import path
from setuptools import setup, find_packages

from hip_pocket.constants import VERSION


def load(file_name):
    here = path.dirname(path.abspath(__file__))
    return open(path.join(here, file_name), "r").read()

setup(
    name="HipPocket",
    description="A wrapper around Flask to ease the development of larger applications",
    long_description=load("README.rst"),
    version=VERSION,
    packages=find_packages(),
    url="https://github.com/svieira/HipPocket",
    author="Sean Vieira",
    author_email="vieira.sean+hip_pocket@gmail.com",
    install_requires=[
        "Flask>=0.7",
        "Jinja2>=2.4",
        "Werkzeug>=0.7"
    ],
    platforms="any",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ],
    include_package_data=True,
    test_suite="hip_pocket.tests.suite",
    zip_safe=False
    )
